//
//  Model.swift
//  MorseTutor
//

import AppKit

class Model  {
    static let minWordsPerMinute = 5
    static let maxWordsPerMinute = 20
    
    fileprivate let PercentReviewTarget = 75
    fileprivate let NewSymbolsTarget = 2
    
    // These three values are dummies so that init can call a function to initialize these three variables
    fileprivate var percentReview: Int = 0
    fileprivate var reviewSymbols: [Symbol] = []
    fileprivate var newSymbols: [Symbol] = []
    
    fileprivate weak var delegate: ModelDelegate?

    fileprivate var timer: Timer!
    fileprivate var timerEndTime: Date!
    fileprivate var timerEnded = false, abortSound = false
    
    // The reason we keep track of spaces as a "symbol tested" and yet subtract it off at the end is so that spaces can be inserted after every group of five symbols output.
    fileprivate var numSymbolsTested = 0, numSpacesOutput = 0, numSymbolsCorrect = 0
    fileprivate(set) var serialCounter = 0
    var percentCorrect:Double { return Double(numSymbolsCorrect) / Double(numSymbolsTested - numSpacesOutput) }

    fileprivate var guessableSymbol: Symbol!, nextSymbol: Symbol!
    var guessableSymbolKeyboardLength:Int { return guessableSymbol != nil ? (guessableSymbol.keyboardInput as NSString).length : 0 }
    
    fileprivate var soundHelper: SoundHelper!
    
    init?(wordsPerMinute: Int, highestSymbol:Int, emphasizeNewSymbols:Bool, delegate del:ModelDelegate) {
        soundHelper = SoundHelper(wordsPerMinute: wordsPerMinute)
        delegate = del

        makeSymbolVectors(highestSymbol, emphasizeNewSymbols: emphasizeNewSymbols)
        delegate?.isGuessable(false)
    }
    
    // When the parameters change, soundHelper and the symbol vectors need to be reinitialized. This works better than freeing and allocating new Model and SoundHelper objects
    func reinit(_ wordsPerMinute: Int, highestSymbol:Int, emphasizeNewSymbols:Bool, delegate del:ModelDelegate) {
        soundHelper.reinit(wordsPerMinute: wordsPerMinute)
        numSymbolsTested = 0
        numSpacesOutput = 0
        numSymbolsCorrect = 0
        delegate = del
        
        guessableSymbol = nil
        nextSymbol = nil
        
        delegate?.isGuessable(false)
        delegate?.guessableNext(guessableSymbol != nil ? guessableSymbol.displayLabel : "", nextSymbol: nextSymbol != nil ? nextSymbol.displayLabel : "")
        
        makeSymbolVectors(highestSymbol, emphasizeNewSymbols: emphasizeNewSymbols)
    }
    
    // Called by init and reinit
    func makeSymbolVectors(_ highestSymbol:Int, emphasizeNewSymbols:Bool) {
        let numSymbols = highestSymbol - Symbol.minG4FONOrdinal + 1
        let numNewSymbols = emphasizeNewSymbols ? min(NewSymbolsTarget, numSymbols) : 0
        let numReviewSymbols = numSymbols - numNewSymbols
        
        
        if emphasizeNewSymbols {
            if numReviewSymbols == 0 {
                // No symbols to review (i.e., all symbols are "new") thus no time is spent on review
                percentReview = 0
            }
            else if numNewSymbols*PercentReviewTarget > (100-PercentReviewTarget)*numReviewSymbols {
                // If numNewSymbols == 2 and numReviewSymbols < 6, then the review symbols will actually be favored because 75% of the time is devoted to review. In this case make all the frequencies the same.
                percentReview = 100*numReviewSymbols/(numReviewSymbols+numNewSymbols)
            }
            else {
                // Spend 75% of time on old symbols and 25% of time on the most recent two symbols
                percentReview = PercentReviewTarget
            }
        }
        else {
            // If the user does not want to emphasize the most recent few symbols, all symbols are treated as review symbols and 100% of time is spent on review (duh).
            percentReview = 100
        }
        
        let lowestReview = Symbol.minG4FONOrdinal
        let highestReview = Symbol.minG4FONOrdinal + numReviewSymbols - 1
        let lowestNew = highestReview + 1
        let highestNew = lowestNew + numNewSymbols - 1
        
        reviewSymbols = lowestReview <= highestReview ? (lowestReview ... highestReview).map { i in Symbol(sequenceG4FON: i)! } : []
        newSymbols = emphasizeNewSymbols ? (lowestNew ... highestNew).map { i in Symbol(sequenceG4FON: i)! } : []
    }

    func startTimer(_ seconds: Int) {
        // Don't have two timers going at once
        if timer != nil {
            return
        }
        
        timerEndTime = Date(timeIntervalSinceNow: TimeInterval(seconds))
        timerEnded = false
        
        // This class needs to inherit from NSObject for this to work
        timer = Timer.scheduledTimer(timeInterval: 1.0, target: self, selector: #selector(timerFireMethod), userInfo: nil, repeats: true)
        
        enqueueRandomSymbol()
    }
    
    // Allows a symbol being played to finish
    func endTimer() {
        // signal the symbol generator to stop. (7 means least-significant bit)
        OSAtomicTestAndSet(7, &timerEnded)
        
        timer.invalidate()
        timer = nil
        
        // Allow for grace period for guessing until endOfExtraTime is called
    }
    
    // Termiminates the audio immediately
    func abortTimer() {
        // signal the symbol generator to stop. (7 means least-significant bit)
        OSAtomicTestAndSet(7, &timerEnded)

        timer.invalidate()
        soundHelper.abortSound()
        timer = nil

        guessableSymbol = nil
        nextSymbol = nil
        
        delegate?.isGuessable(false)
        delegate?.guessableNext(guessableSymbol != nil ? guessableSymbol.displayLabel : "", nextSymbol: nextSymbol != nil ? nextSymbol.displayLabel : "")
    }
    
    @objc func timerFireMethod(_ timer: Timer) {
        let now = Date(timeIntervalSinceNow: 0)
        
        // if now is later than or equal to timerEndTime, i.e. we have reached the scheduled end of the timer
        if !(now.compare(timerEndTime) == .orderedAscending) {
            endTimer()
            delegate?.timerExpired()
        }
        // else, we have not reached the scheduled end of the timer
        else {
            let doubleSecondsLeft = timerEndTime.timeIntervalSince(now)
            let intSecondsLeft = Int(round(doubleSecondsLeft))
            delegate?.timerTick(intSecondsLeft)
        }
    }
    
    // User allowed to guess a symbol up until the next symbol is done being played
    // Must be called from main thread
    func enqueueRandomSymbol() {
        assert(Thread.isMainThread)
    
        // If timer is over (or aborted) stop queuing more random symbols
        if timerEnded {
            // "guess" can be called after the timer is expired to guess the last symbol
            // either both are nil already or guessableSymbol is already the just-played symbol (e.g., due to an earlier guess), in which case, do nothing
            if guessableSymbol != nil && nextSymbol != nil {
                guessableSymbol = nextSymbol
                nextSymbol = nil

                delegate?.isGuessable(guessableSymbol != nil)
                delegate?.guessableNext(guessableSymbol != nil ? guessableSymbol.displayLabel : "", nextSymbol: nextSymbol != nil ? nextSymbol.displayLabel : "")
            }
            
            return
        }

        var randomSymbol: Symbol!
        
        if (numSymbolsTested + 1) % 6 == 0 {
            // After every five symbols output a space
            randomSymbol = Symbol(space: " ")
        }
        else {
            // Output a "review" symbol (i.e., any but most recent two symbols) reviewPercent of the time
            if Int(arc4random_uniform(100) + 1) <= self.percentReview {
                // Review old symbol
                randomSymbol = reviewSymbols[Int(arc4random_uniform(UInt32(reviewSymbols.count)))]
            }
            else {
                // New symbol
                randomSymbol = newSymbols[Int(arc4random_uniform(UInt32(newSymbols.count)))]
            }
        }
        
        // Comments refer to symbols A (two symbols ago), B (symbol that just finished playing) and C (randomSymbol, the one just chosen which has not yet started playing). This assumes that C is the third or later symbol randomly chosen, if C is the first or second symbol randomly chosen the comments refer to A and/or B "not existing"
        
        if guessableSymbol == nil {
            // (i) C is the first symbol randomly selected (A and B don't exist) or (ii) the user already attempted to guess B (which means A was also attempted or A does not exist). The user has the chance to guess C.
            
            // This situation does not represent a "failure to guess" in either (i) or (ii) so do not call delegate.printWrong
            
            guessableSymbol = randomSymbol
            assert (nextSymbol == nil)
        }
        else {
            if nextSymbol == nil {
                // (i) User attempted to guess A before B finished playing (thus making B the guessableSymbol and nextSymbol nil), but has not attempted to guess B (thus the guessableSymbol is still B), or (ii) A doesn't exist. In this situation the user still has the opportunity to guess B until C is finished playing.
                
                // This situation does not represent a "failure to guess" in either (i) or (ii) so do not call delegate.printWrong
                
                nextSymbol = randomSymbol
            }
            else {
                // User did not attempt to guess A before B finished playing (thus making A the guessableSymbol and B the nextSymbol). The user has the chance to guess B until C is finished playing.
                
                // This situation represents a "failure to guess" therefore call delegate.printWrong
                delegate?.printWrong(guessableSymbol.displayLabel)
                
                guessableSymbol = nextSymbol
                nextSymbol = randomSymbol
            }
        }
        
        soundHelper.enqueueSymbol(randomSymbol, completionHandler: { DispatchQueue.main.async {
            self.enqueueRandomSymbol() } } )
        numSymbolsTested += 1
        numSpacesOutput += randomSymbol.keyboardInput == " " ? 1 : 0
        serialCounter += 1

        delegate?.isGuessable(guessableSymbol != nil)
        delegate?.guessableNext(guessableSymbol != nil ? guessableSymbol.displayLabel : "", nextSymbol: nextSymbol != nil ? nextSymbol.displayLabel : "")
        
        // Not practical for user to manually guess spaces, because user needs to wait until the symbol is fully played to guess
        if guessableSymbol?.keyboardInput == " " {
            guess(" ")
        }
    }
    
    // Must be called from main thread
    func guess(_ keyboardText: String) {
        assert(Thread.isMainThread)

        // If nil, then symbol has already been guessed and printCorrect/Wrong already called
        guard guessableSymbol != nil else {
            return
        }
        
        if guessableSymbol.keyboardInput == keyboardText {
            delegate?.printCorrect(guessableSymbol.displayLabel)
            
            // Don't count spaces as correct because those are automatically guessed
            numSymbolsCorrect += guessableSymbol.keyboardInput == " " ? 0 : 1
        } else {
            delegate?.printWrong(guessableSymbol.displayLabel)
        }
        
        // Comments refer to symbols B (symbol that just finished playing) and C (the symbol currently playing). This assumes C is the second or later symbol randomly chosen, if C is the first symbol the comments refer to B "not existing."
        
        // If nextSymbol is not nil, then guessableSymbol (the symbol just guessed) is B and nextSymbol is C. The user now has the chance to guess C. If nextSymbol is nil, then guessableSymbol (the symbol just guessed) is C and guesses will be ignored until the next symbol starts to play.
        guessableSymbol = nextSymbol
        nextSymbol = nil
        
        delegate?.isGuessable(guessableSymbol != nil)
        delegate?.guessableNext(guessableSymbol != nil ? guessableSymbol.displayLabel : "", nextSymbol: nextSymbol != nil ? nextSymbol.displayLabel : "")
        
        // Not practical for user to manually guess spaces, because user needs to wait until the symbol is fully played to guess
        if guessableSymbol?.keyboardInput == " " {
            guess(" ")
        }
    }
    
    // After the timer is up, the controller can allow the user to guess the symbol that was in progress. After some grace period (e.g. 1-2 seconds) the controller should no longer allow the user to guess. Controller should call this function to count that symbol as a wrong guess if the user did not previously guess that symbol during the grace period.
    func endOfExtraTime() {
        if guessableSymbol != nil {
            delegate?.printWrong(guessableSymbol.displayLabel)
            
            guessableSymbol = nil
            nextSymbol = nil

            delegate?.isGuessable(false)
            delegate?.guessableNext(guessableSymbol != nil ? guessableSymbol.displayLabel : "", nextSymbol: nextSymbol != nil ? nextSymbol.displayLabel : "")
        }
    }
}

protocol ModelDelegate : class {
    func timerTick (_ secondsLeft: Int)
    func timerExpired ()
    func printCorrect (_ symbolText: String)
    func printWrong (_ symbolText: String)
    func isGuessable (_ guessable: Bool)
    func guessableNext (_ guessableSymbol: String, nextSymbol: String)
}
