//
//  SoundHelper.swift
//  MorseTutor
//

import Foundation
import AVFoundation

class SoundHelper {
    fileprivate let sampleRate = 44100

    fileprivate var wordsPerMinute: Int
    fileprivate var audioEngine: AVAudioEngine
    fileprivate var audioPlayerNode: AVAudioPlayerNode
    fileprivate var audioFormat: AVAudioFormat

    fileprivate lazy var ditBuffer:AVAudioPCMBuffer = self.allocDitBuffer()
    fileprivate lazy var dahBuffer:AVAudioPCMBuffer = self.allocDahBuffer()
    fileprivate lazy var interSymbolBuffer:AVAudioPCMBuffer = self.allocInterSymbolBuffer()
    fileprivate lazy var spaceBuffer:AVAudioPCMBuffer = self.allocSpaceBuffer()
    
    init (wordsPerMinute wpm: Int) {
        wordsPerMinute = wpm
        
        audioEngine = AVAudioEngine()
        audioPlayerNode = AVAudioPlayerNode()
        audioFormat = AVAudioFormat(commonFormat: .pcmFormatFloat32, sampleRate: Double(sampleRate), channels: 2, interleaved: false)!
        let mainMixer = audioEngine.mainMixerNode

        audioEngine.attach(audioPlayerNode)
        audioEngine.connect(audioPlayerNode, to:mainMixer, format: audioFormat)
        try! audioEngine.start()
        audioPlayerNode.play()
    }

    func reinit (wordsPerMinute wpm: Int) {
        wordsPerMinute = wpm
        
        ditBuffer = allocDitBuffer()
        dahBuffer = allocDahBuffer()
        interSymbolBuffer = allocInterSymbolBuffer()
        spaceBuffer = allocSpaceBuffer()

        audioPlayerNode.play()
    }

    
    func abortSound() {
        audioPlayerNode.stop()
    }
    
    // AARL code practice transmission = 750 Hz
    fileprivate let hertz = 750.0
    
    // Farnsworth timing based on AARL code practice transmission @ 20 wpm
    // All constants in seconds
    fileprivate let rise = 0.005, ditSustain = 0.055, dahSustain = 0.175, fall = 0.010
    fileprivate let intraSymbol = 0.050
    
    // The interSymbol and space timing is scaled up to achieve the desired wpm (for wpm < 20)
    fileprivate let interSymbol20 = 0.170, space20 = 0.410
    
    fileprivate func farnsworthScaleFactor() -> Double {
        // In standard word "PARIS " there are 10 dits, 4 dahs, 9 brief intra-symbol silences between dits and dahs, 4 silences between symbols and one space
        return (60.0/Double(wordsPerMinute)-(10*(rise+ditSustain+fall)+4*(rise+dahSustain+fall)+9*intraSymbol))/(4*(interSymbol20)+space20)
    }
    
    // after each symbol, in seconds
    fileprivate var interSymbol:Double {
        return interSymbol20 * farnsworthScaleFactor()
    }
    
    // between words, in seconds
    fileprivate var space:Double {
        return space20 * farnsworthScaleFactor()
    }

    fileprivate func linear_interpolate (xt:Double, x1:Double, y1:Double, x2:Double, y2:Double) -> Double {
        return y1 + (xt - x1)*(y2 - y1)/(x2 - x1)
    }
    
    // Some hams consider the "raised sine" to be the ideal CW waveform. for x = 0, returns 0. for x = 1, return 1. interpolates in-between in a sinusoidal way.
    fileprivate func raised_sine (_ x:Double) -> Double {
        return sin((x-0.5)*Double.pi)*0.5+0.5
    }
    
    fileprivate func allocDitBuffer() -> AVAudioPCMBuffer {
        let ditLength = rise + ditSustain + fall + intraSymbol
        let ditFrameCount = Int(Double(sampleRate) * ditLength)
        let ditBuffer = AVAudioPCMBuffer(pcmFormat: audioFormat, frameCapacity: UInt32(ditFrameCount))!
        ditBuffer.frameLength = ditBuffer.frameCapacity
        
        for i in 0 ..< ditFrameCount {
            let t = Double(i)/Double(sampleRate)
            let sine = sin(t*2.0*Double.pi*hertz)
            
            var shape: Double
            if t < rise {
                shape = raised_sine(linear_interpolate(xt:t, x1: 0.0, y1: 0.0, x2: rise, y2: 1.0))
            } else if t < (rise + ditSustain) {
                shape = 1.0
            } else if t < (rise + ditSustain + fall) {
                shape = raised_sine(linear_interpolate(xt:t, x1: rise + ditSustain, y1: 1.0, x2: rise + ditSustain + fall, y2: 0.0))
            } else {
                shape = 0.0
            }
            
            ditBuffer.floatChannelData?[0][i] = Float(sine * shape)
            ditBuffer.floatChannelData?[1][i] = Float(sine * shape)
        }

        return ditBuffer
    }

    fileprivate func allocDahBuffer() -> AVAudioPCMBuffer {
        let dahLength = rise + dahSustain + fall + intraSymbol
        let dahFrameCount = Int(Double(sampleRate) * dahLength)
        let dahBuffer = AVAudioPCMBuffer(pcmFormat: audioFormat, frameCapacity: UInt32(dahFrameCount))!
        dahBuffer.frameLength = dahBuffer.frameCapacity
        
        for i in 0 ..< dahFrameCount {
            let t = Double(i)/Double(sampleRate)
            let sine = sin(t*2.0*Double.pi*hertz)
            
            var shape: Double
            if t < rise {
                shape = raised_sine(linear_interpolate(xt: t, x1: 0.0, y1: 0.0, x2: rise, y2: 1.0))
            } else if t < (rise + dahSustain) {
                shape = 1.0
            } else if t < (rise + dahSustain + fall) {
                shape = raised_sine(linear_interpolate(xt: t, x1: rise + dahSustain, y1: 1.0, x2: rise + dahSustain + fall, y2: 0.0))
            } else {
                shape = 0.0
            }
            
            dahBuffer.floatChannelData?[0][i] = Float(sine * shape)
            dahBuffer.floatChannelData?[1][i] = Float(sine * shape)
        }
        
        return dahBuffer
    }
    
    fileprivate func allocInterSymbolBuffer() -> AVAudioPCMBuffer {
        // the dit and dah buffers already have an intra-symbol space at the end
        let interSymbolFrameCount = Int(Double(sampleRate) * (interSymbol - intraSymbol))
        let interSymbolBuffer = AVAudioPCMBuffer(pcmFormat: audioFormat, frameCapacity: UInt32(interSymbolFrameCount))!
        interSymbolBuffer.frameLength = interSymbolBuffer.frameCapacity
        
        bzero(interSymbolBuffer.floatChannelData?[0], MemoryLayout<Float>.size * interSymbolFrameCount)
        bzero(interSymbolBuffer.floatChannelData?[1], MemoryLayout<Float>.size * interSymbolFrameCount)
        
        return interSymbolBuffer
    }
    
    fileprivate func allocSpaceBuffer() -> AVAudioPCMBuffer {
        // after a symbol is queued an intersymbol space is automatically queued. of course the timings only work right if spaces only follow symbols (i.e., no double spaces)
        let spaceFrameCount = Int(Double(sampleRate) * (space - interSymbol))
        let spaceBuffer = AVAudioPCMBuffer(pcmFormat: audioFormat, frameCapacity: UInt32(spaceFrameCount))!
        spaceBuffer.frameLength = spaceBuffer.frameCapacity
        
        bzero(spaceBuffer.floatChannelData?[0], MemoryLayout<Float>.size * spaceFrameCount)
        bzero(spaceBuffer.floatChannelData?[1], MemoryLayout<Float>.size * spaceFrameCount)

        return spaceBuffer
    }
    
    func enqueueSymbol(_ symbol: Symbol, completionHandler:@escaping () -> Void) {
        if symbol.ditDah == " " {
            audioPlayerNode.scheduleBuffer(spaceBuffer, at: nil, options: [], completionHandler: { completionHandler() } )
            return
        }
        
        for d in symbol.ditDah {
            if d == "." {
                audioPlayerNode.scheduleBuffer(ditBuffer, at: nil, options: [], completionHandler: nil)
            } else if d == "-" {
                audioPlayerNode.scheduleBuffer(dahBuffer, at: nil, options: [], completionHandler: nil)
            }
        }
        
        audioPlayerNode.scheduleBuffer(interSymbolBuffer, at: nil, options: [], completionHandler: { completionHandler() } )
    }
}
