//
//  ViewController.swift
//  MorseTutor
//

import Cocoa

class ViewController: NSViewController, ModelDelegate {
    @IBOutlet weak var textArea: NSTextField!

    @IBOutlet weak var wpmSlider: NSSlider!
    @IBOutlet weak var wpmTextField: NSTextField!
    
    @IBOutlet weak var highestSlider: NSSlider!
    @IBOutlet weak var highestTextField: NSTextField!
    
    @IBOutlet weak var statusTextField: NSTextField!
    @IBOutlet weak var guessableIndicator: NSTextField!
    @IBOutlet weak var reviewRecentCheckbox: NSButton!
    @IBOutlet weak var startOrStopButton: NSButton!
    
    @IBOutlet weak var guessableSymbolLabel: NSTextField!
    @IBOutlet weak var nextSymbolLabel: NSTextField!
    
    @objc dynamic var wpmSliderValue:Double = 0.0 { didSet { wpmTextField.stringValue = "\(wordsPerMinute)" } }
    var wordsPerMinute:Int { return Int(wpmSliderValue) }
    
    @objc dynamic var highestSliderValue:Double = 0.0 { didSet { highestTextField.stringValue = Symbol(sequenceG4FON: highestSymbol)!.displayLabel } }
    var highestSymbol:Int { return Int(highestSliderValue) }

    var model:Model!
    var started:Bool = false
    let timerSeconds = 300

    // Give user a little time to key in the last symbol when time expires
    let extraSeconds = 2
    var extraTimer:Timer!
    
    // Some guesses (prosigns) can be multi-character, this string accumulates characters until the required number of characters is reached
    var guess:String = ""
    var serialCounter = 0
    
    struct defaultsConstants {
        static let wpmSliderValue = "wpmSliderValue"
        static let highestSliderValue = "highestSliderValue"
        static let reviewRecentCheckbox = "reviewRecentCheckbox"
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()

        wpmSlider.minValue = Double(Model.minWordsPerMinute)
        wpmSlider.maxValue = Double(Model.maxWordsPerMinute)
        wpmSlider.numberOfTickMarks = Model.maxWordsPerMinute - Model.minWordsPerMinute + 1
        highestSlider.minValue = Double(Symbol.minG4FONOrdinal)
        highestSlider.maxValue = Double(Symbol.maxG4FONOrdinal)
        highestSlider.numberOfTickMarks = Symbol.maxG4FONOrdinal - Symbol.minG4FONOrdinal + 1

        let defaults = UserDefaults.standard
        wpmSliderValue = defaults.object(forKey: defaultsConstants.wpmSliderValue) != nil ? defaults.double(forKey: defaultsConstants.wpmSliderValue) : wpmSlider.minValue
        highestSliderValue = defaults.object(forKey: defaultsConstants.highestSliderValue) != nil ? defaults.double(forKey: defaultsConstants.highestSliderValue) : highestSlider.minValue
        reviewRecentCheckbox.state = defaults.object(forKey: defaultsConstants.reviewRecentCheckbox) != nil ? NSControl.StateValue(rawValue: defaults.integer(forKey: defaultsConstants.reviewRecentCheckbox)) : NSControl.StateValue.on

        model = Model(wordsPerMinute: wordsPerMinute, highestSymbol: highestSymbol, emphasizeNewSymbols: (reviewRecentCheckbox.state == NSControl.StateValue.on), delegate: self)
        
        guessableSymbolLabel.stringValue = ""
        nextSymbolLabel.stringValue = ""
    }

    override func viewWillDisappear() {
        super.viewWillDisappear()
        
        let defaults = UserDefaults.standard
        defaults.set(wpmSliderValue, forKey: defaultsConstants.wpmSliderValue)
        defaults.set(highestSliderValue, forKey: defaultsConstants.highestSliderValue)
        defaults.set(reviewRecentCheckbox.state, forKey: defaultsConstants.reviewRecentCheckbox)
    }

    struct labels {
        static let start = "Start"
        static let stop = "Stop"
        static let countdown = "Time remaining: %02d:%02d"
        static let results = "Percent accurate: %.f%%"
    }
    
    @IBAction func startOrStop(_ sender: NSButton) {
        if !started {
            textArea.stringValue = ""
            
            statusTextField.stringValue = String(format:labels.countdown, timerSeconds / 60, timerSeconds % 60)
            statusTextField.isHidden = false
            
            sender.title = labels.stop
            wpmSlider.isEnabled = false
            highestSlider.isEnabled = false
            reviewRecentCheckbox.isEnabled = false

            model.reinit(wordsPerMinute, highestSymbol: highestSymbol, emphasizeNewSymbols: (reviewRecentCheckbox.state == NSControl.StateValue.on), delegate: self)
            model.startTimer(timerSeconds)
            
            started = true
        }
        else {
            model.abortTimer()
            if extraTimer != nil {
                extraTimer.invalidate()
                extraTimer = nil
            }
            
            statusTextField.stringValue = String(format:labels.results, model.percentCorrect * 100)
            
            sender.title = labels.start
            wpmSlider.isEnabled = true
            highestSlider.isEnabled = true
            reviewRecentCheckbox.isEnabled = true
            
            started = false
            guessableIndicator.isHidden = true
        }
    }

    func timerTick (_ secondsLeft: Int) {
        statusTextField.stringValue = String(format:labels.countdown, secondsLeft / 60, secondsLeft % 60)
    }
    
    func timerExpired () {
        statusTextField.stringValue = String(format:labels.countdown, 0, 0)

        extraTimer = Timer.scheduledTimer(timeInterval: 2.0, target: self, selector: #selector(extraTimerExpired), userInfo: nil, repeats: false)
    }
    
    @objc func extraTimerExpired () {
        extraTimer = nil
        
        statusTextField.stringValue = String(format:labels.results, model.percentCorrect * 100)
        
        startOrStopButton.title = labels.start
        wpmSlider.isEnabled = true
        highestSlider.isEnabled = true
        reviewRecentCheckbox.isEnabled = true
        
        started = false
    }
    
    func printCorrect (_ symbolText: String) {
        let attributedSymbolText = NSAttributedString(string: symbolText, attributes: [NSAttributedString.Key.foregroundColor: NSColor.textColor])

        let newText = textArea.attributedStringValue.mutableCopy() as! NSMutableAttributedString
        newText.append(attributedSymbolText)
        textArea.attributedStringValue = newText
        guess = ""
    }
    
    func printWrong (_ symbolText: String) {
        let attributedSymbolText = NSAttributedString(string: symbolText, attributes: [NSAttributedString.Key.foregroundColor: NSColor.red])
        
        let newText = textArea.attributedStringValue.mutableCopy() as! NSMutableAttributedString
        newText.append(attributedSymbolText)
        textArea.attributedStringValue = newText
        guess = ""
    }
    
    func isGuessable(_ guessable: Bool) {
        // The indicator is red, so show it when the lesson has started and when the symbol is not guessable
        guessableIndicator.isHidden = !started || guessable
        
    }
    
    func guessableNext(_ guessableSymbol: String, nextSymbol: String) {
        guessableSymbolLabel.stringValue = guessableSymbol
        nextSymbolLabel.stringValue = nextSymbol
    }
    
    // Called from WindowController to handle keystrokes. Returns true if the keystroke is handled here, false otherwise
    func handleKeyStroke (_ theEvent: NSEvent) -> Bool {
        guard theEvent.modifierFlags.intersection([.control, .option, .command, .help, .function]) == [] && theEvent.characters != nil else {
            return false
        }
        
        let validChars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789.,?/"
        if validChars.contains(theEvent.characters!) {
            let key = theEvent.characters!.uppercased()
            
            // Model's serialCounter is also initialized to 0. This makes sure that multicharacter guesses are applied to the right symbol. E.g., one character of a prosign is typed, then the guessable symbol changes, and a second character is typed - the second character should not be applied to the prosign, which is no longer guessable
            if serialCounter != model.serialCounter {
                guess = ""
                serialCounter = model.serialCounter
            }

            if model.guessableSymbolKeyboardLength < ((guess as NSString).length + 1) {
                // Too many characters (which probably means the user already guessed, so that guessableSymbol is nil thus guessableSymbolKeyboardLength is 0), so let WindowController's superclass handle (i.e., beep at the user)
                
                return false
            }
            
            guess.append(key)
            if model.guessableSymbolKeyboardLength == (guess as NSString).length {
                model.guess(guess)
            }
            
            return true
        }
        
        // Some other strange character was pressed, so let WindowController's superclass handle (i.e., beep at the user)
        return false
    }
    
}

